/* aica.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ok_rand.h"
#include "ok_asn1.h"
#include "ok_x509.h"
#include "ok_pkcs.h"
#include "ok_pem.h"
#include "ok_tool.h"
#include "ok_uconv.h"

#include "ok_aica.h"

#ifndef	AICONFIG
#define	AICONFIG "aica.cnf"
#endif

enum {PF_DEL,PF_ADD};

char cap12[256] = "";

char reqf[256]  = "";
char resign[256]= "";
char pfname[32] = "SMIME user";
char csvf[256]  = "";
char outf[256]  = "newcert.cer";
char conf[256]  = AICONFIG;

long revoke=0;
int revrsn=0;
int keysize=1024;
int keytype=KEY_RSA_PUB;
int days=365,sn=0,pf_all=0;
int caop=0,caext=0;
int pfmode = PF_ADD;

/* ca_conf.c */
extern char ca_path[]; /* open ca path */

/* other functions */
int CA_do_operation(CA *ca);
int ca_set_defdir(CertDN *dn);

int req_config(char *fname);
void options(int argc,char **argv);
void usage(int mode);


/*-------------------------------------------------
	begin aica main
-------------------------------------------------*/
int main(int argc,char **argv){
	int i,err=-1,change=0;
	char path[256];
	CA *ca=NULL;
	CertProf *cpf;

	*cap12=*reqf=*resign=0;

	OK_clear_error();
	options(argc,argv);
	RAND_init();

#ifdef __WINDOWS__
	get_aiconf_path(conf,256);
#endif

	if(CA_read_config(conf)){
		printf("cannot read a config file : %s\n",conf);
		goto done;
	}

	/*
	 * a new CA generation mode. -- the new CA will be created
	 * with new key pairs and a certificate. Also it is possible
	 * to use a PKCS#12 file as a CA certificate and key pairs.
	 */
	if(caop == AICA_OP_NEW){
		if((ca = CA_new())==NULL) goto done;
		if(CA_set_ca_path(ca,ca_path)) goto done;
		if(CA_set_lib_path(ca)) goto done;
		if(req_config(conf)) goto done;

		if(*cap12){ /* use other PKCS#12 file */
			if(i=CA_new_with_p12(ca)){
				if(i<0) printf("CA Certificate & Private Key are not found.\n");
				else printf("PKCS#12 file does not contain valid CA certificate.\n");
				return -1;
			}

		}else{	  /* generage new CA cert & key */
			if(CA_new_generate(ca,keysize,keytype)) goto done;

			/* set CA certificate validation */
			if(days){ ca->cprof->cert_days=days;}

			if(Sign_do(ca,ca->cert)) goto done;
			
			/* clean temporary ca cert template */
			cpf = ca->profList;
			ca->profList = cpf->next;
			Prof_cert_free(cpf); cpf=NULL;
		}

		/* update and write files */
		if(CA_update_P12andInfo(ca,*cap12)) goto done;

		if((ca->der = CA_toDER(ca,NULL,&i))==NULL) goto done;
		if(CA_info_write(ca)) goto done;

		err=0; goto done; /* exit(0) */
	}
	
	/*
	 * normal CA booting
	 */
	if((ca=CA_info_read(ca_path))==NULL){
		printf("CA Information not found.\n");
		goto done;
	}
	if(CA_set_lib_path(ca)) goto done;
	if(CA_p12_read(ca)){
		printf("CA Certificate & Private Key are not found.\n");
		goto done;
	}
	if(CA_expire_check(ca)) change++;

	/* these functions must be called after CA_info_read() */
	/* open certificate profiles. */
	strncpy(path,ca_path,210);
	strcat (path,PATH_DELI);
	strcat (path,"cert");

	for(ca->cprof=ca->profList; ca->cprof; ca->cprof=ca->cprof->next){
		if(Prof_open_ctinfo(ca,path)){
			/* create New Profile Info (for version compatibility) */
			if(Prof_save_ctinfo(ca,path)) goto done;
		}
		if(Prof_open_ctfile(ca,path)) goto done;
	}

	for(ca->lprof=ca->crlpList; ca->lprof; ca->lprof=ca->lprof->next){
		if(Prof_open_clinfo(ca,path)){
			/* create New Profile Info (for version compatibility) */
			if(Prof_save_clinfo(ca,path)) goto done;
		}
	}
	/* set current certificate profile */
	ca->cprof = Prof_find(ca,pfname);
	ca->lprof = Prof_crl_find(ca,pfname);
	if((ca->cprof==NULL)&&(ca->lprof==NULL)){
		printf("cannot set current profile : unknown profile name.\n");
		goto done;
	}

	/* 
	 * do operation
	 */
	if((change=CA_do_operation(ca))<0) goto done;

	if(change){
		if(ca->der) AIFREE(ca->der);
		if((ca->der = CA_toDER(ca,NULL,&i))==NULL) goto done;
		if(CA_info_write(ca)) goto done;

		for(ca->cprof=ca->profList; ca->cprof; ca->cprof=ca->cprof->next)
			if(Prof_save_ctinfo(ca,path)) goto done;

		for(ca->lprof=ca->crlpList; ca->lprof; ca->lprof=ca->lprof->next)
			if(Prof_save_clinfo(ca,path)) goto done;
	}

	err=0;
done:
	RAND_cleanup();
	free_u2j_table();

	CA_free(ca);
	if(err) printf("%s\n",CA_get_errstr());
	return err;
}

/*-----------------------------------------
  usage and option check
-----------------------------------------*/
void options(int argc,char **argv){
	int i=1;

#ifdef __WINDOWS__
	OK_set_pwd_dosmode(1);
#endif
	if(argc<=1){
		usage(0);
		exit(0);
	}

	/* check operation first */
	if(!strcmp("sign",argv[i])){
		caop = AICA_OP_SIGN;

	}else if(!strcmp("resign",argv[i])){
		caop = AICA_OP_RESIGN;

	}else if(!strcmp("crl",argv[i])){
		caop = AICA_OP_CRL;

	}else if(!strcmp("revoke",argv[i])){
		caop = AICA_OP_REVOKE;
		i++;
		if(i<argc) revoke=atoi(argv[i]);
	}else if(!strcmp("csv",argv[i])){
		caop = AICA_OP_CSV;
		i++;
		if(i<argc) strncpy(csvf,argv[i],254);
	}else if(!strcmp("print",argv[i])){
		caop = AICA_OP_PRINT;

	}else if(!strcmp("list",argv[i])){
		caop = AICA_OP_LIST;

	}else if(!strcmp("new",argv[i])){
		caop = AICA_OP_NEW;

	}else if(!strcmp("pfset",argv[i])){
		caop = AICA_OP_PFSET;

	}else if(!strcmp("pfext",argv[i])){
		caop = AICA_OP_PFEXT;

	}else if(!strcmp("prof",argv[i])){
		caop = AICA_OP_PROF;
	
	}else{
		printf("operation error!\n");
		usage(0);
		exit(0);
	}

	/* check options */
	for(i++;i<argc;i++){
		if(!strcmp("-conf",argv[i])){
			i++;
			if(i<argc) strncpy(conf,argv[i],254);
		}else if(!strcmp("-pf",argv[i])){
			i++;
			if(i<argc) strncpy(pfname,argv[i],30);
		}else if(!strcmp("-sn",argv[i])){
			i++;
			if(i<argc) sn=atoi(argv[i]);
		}else if(!strcmp("-o",argv[i])){
			i++;
			if(i<argc) strncpy(outf,argv[i],254);
		}

		/* options for profile add or delete */
		else if(!strcmp("-add",argv[i])){
			pfmode = PF_ADD;

		}else if(!strcmp("-del",argv[i])){
			pfmode = PF_DEL;
		}

		/* options for CA generation */
		else if(!strcmp("-ksize",argv[i])){
			i++;
			if(i<argc) keysize=atoi(argv[i]);

		}else if(!strcmp("-ktype",argv[i])){
			/* nothing to do now */

		}else if(!strcmp("-cap12",argv[i])){
			i++;
			if(i<argc) strncpy(cap12,argv[i],254);
		}

		/* options for profile settings */
		else if(!strcmp("-days",argv[i])){
			i++;
			if(i<argc) days=atoi(argv[i]);
		}
		else if(!strcmp("-all",argv[i])){
			pf_all++;
		}
		
		/* options for profile extensions */		
		else if(!strcmp("-nscrl",argv[i])){
			caext = OBJ_NS_CERT_CRLURL;

		}else if(!strcmp("-nscomm",argv[i])){
			caext = OBJ_NS_CERT_COMMENT;

		}else if(!strcmp("-nstype",argv[i])){
			caext = OBJ_NS_CERT_TYPE;

		}else if(!strcmp("-keyusage",argv[i])){
			caext = OBJ_X509v3_KEY_Usage;

		}else if(!strcmp("-bscons",argv[i])){
			caext = OBJ_X509v3_BASIC;

		}else if(!strcmp("-authkeyid",argv[i])){
			caext = OBJ_X509v3_AuthKeyIdt;

		}else if(!strcmp("-sbjkeyid",argv[i])){
			caext = OBJ_X509v3_SbjKeyIdt;

		}else{
			if(argv[i][0] == '-'){
				printf("option error!\n");
				usage(caop);
				exit(0);
			}else if(caop == AICA_OP_SIGN){
				strncpy(reqf,argv[i],254);
			}else if(caop == AICA_OP_RESIGN){
				strncpy(resign,argv[i],254);
			}
		}
	}
}

void usage(int mode){
	if(mode==0){
		printf("usage: aica [operation] [outions]\n");
		printf("operation list\n");
		printf("  sign [file]  : sign on a request \"file\".\n");
		printf("  resign [file]: resign on a certificate \"file\".\n");
		printf("  crl          : output CRL files (CRL,ARL,CRL-ALL).\n");
		printf("  print        : print current CA Information.\n");
		printf("  list         : print issued certificate list.\n");
		printf("  revoke [num] : revoke a certificate with serial \"num\".\n");
		printf("  csv [file]   : issue a lot of certificates with csv \"file\".\n");
		printf("  new          : generate a new CA (ca.p12 & ca.cai file).\n");
		printf("  pfset        : modify certificate or CRL profile settings.\n");
		printf("  pfext        : modify certificate or CRL extensions.\n");
		printf("  prof         : add/delete certificate profile.\n");
		printf("\noptions (general)\n");
		printf("  -conf path   : set aica config path\n");
		printf("\nTo see operation's option list, input 'aica [operation] -help'\n");
		printf("To perform CA operation, input like following example.\n");
		printf("(issuing a new certificate from a request file)\n");
		printf("   aica sign newreq.p10 -sn 10012 -o user01.cer\n");
	}else if(mode){
		printf("usage: aica [operation] [outions]\n");
		printf("specific operation's option\n");
		switch(mode){
		case AICA_OP_SIGN:
			printf("  aica sign [options] file : sign on a request \"file\".\n");
			printf("    -sn num    : set a serial number for signing.\n");
			printf("    -pf name   : select a profile for signing. \n");
			printf("    -o file    : output a certificate named \"file\".\n");
			break;
		case AICA_OP_RESIGN:
			printf("  aica resign [options] file : resign on a certificate \"file\".\n");
			printf("    -o file    : output a certificate named \"file\".\n");
			break;
		case AICA_OP_PRINT:
			printf("  aica print [options] : print current CA Information.\n");
			printf("    -pf name   : select profile to show profile-info. \n");
			printf("    -all       : show all of CA profile-info.\n");
			break;
		case AICA_OP_LIST:
			printf("  aica list [options] : print issued certificate list.\n");
			printf("    -pf name   : select profile for cert-list. \n");
			break;
		case AICA_OP_PFSET:
			printf("  aica pfset [options] : set profile settings.\n");
			printf("    -pf name     : select profile to modify settings.\n");
			printf("    -days num    : set certificate or CRL validity.\n");
			break;
		case AICA_OP_PFEXT:
			printf("  aica pfext [options] : set certificate or CRL extensions.\n");
			printf("    -pf name     : select profile to modify extensions.\n");
			printf("    -bscons      : set Basic Constraints extension.\n");
			printf("    -keyusage    : set Key Usage extension.\n");
			printf("    -authkeyid   : set Authority Key Identifier extension.\n");
			printf("    -sbjkeyid    : set Subject Key Identifier extension.\n");
			printf("    -nscrl       : set Netscape crl URL extension.\n");
			printf("    -nscomm      : set Netscape comment extension.\n");
			printf("    -nstype      : set Netscape cert type extension.\n");
			break;
		case AICA_OP_PROF:
			printf("  aica prof [options] : add/delete certificate profile.\n");
			printf("    -add         : add certificate profile.\n");
			printf("    -del         : delete certificate profile.\n");
			break;
		default:
			printf("  no option for this command.\n");
		}
	}
}

int CA_do_operation(CA *ca){
	CRLProf *lpf = NULL;
	Cert *req = NULL;
	char path[256];
	int ret = -1;

	switch(caop){
	case AICA_OP_PRINT:
		CA_print(ca,pfname,pf_all);
		break;
	case AICA_OP_LIST:
		CA_certstat_print(ca,pfname);
		break;
	case AICA_OP_PFSET:
		if(ca->cprof){ ca->cprof->cert_days = days; ret++;}
		if(ca->lprof){ ca->lprof->crl_days  = days; ret++;}
		break;
	case AICA_OP_PFEXT:
		if((ret=CA_set_extensions(ca,caext))<0) goto done;
		ret--;
		break;
	case AICA_OP_PROF:
		if(CA_do_profile(ca,pfmode)) goto done;
		break;

	case AICA_OP_SIGN:
		/* read and check a request file */
		if((req = Req_read_file(reqf))==NULL){
			printf("Cannot open Certificate request.\n");
			goto done;
		}
		if(Req_signature_verify(req)){
			printf("Certificate request Signature ERROR!\n");
			goto done;
		}
		printf("Certificate Request signature ok.\n");

		/* sign on a request */
		if(sn){
			if(CA_can_serialNum(ca,req,sn)){
				printf("Cannot use serial number : %d\n",sn);
				goto done;
			}
		}
		if(ca_set_defdir(&ca->cert->subject_dn)) goto done;
		if(Sign_do(ca,req)) goto done;
		if(Prof_add_certfile(ca->cprof,req)) goto done;
		if(CA_write_cert(ca,req,outf)) goto done;
		ret++;
		break;
	case AICA_OP_RESIGN:
		if((req = Cert_read_file(resign))==NULL){
			printf("Cannot open a certificate file.\n");
			goto done;
		}
		if(CA_can_resign(ca,req)){
			printf("Cannot resign with this certificate.\n");
			goto done;
		}
		if(Sign_do(ca,req)) goto done;
		if(Prof_add_certfile(ca->cprof,req)) goto done;
		if(CA_write_cert(ca,req,outf)) goto done;
		ret++;
		break;
	case AICA_OP_REVOKE:
		if(CA_do_revocation(ca,revoke)) goto done;
		ret++;
		break;
	case AICA_OP_CRL:
		for(lpf=ca->crlpList; lpf ; lpf=lpf->next){
			ca->lprof = lpf;
			strncpy(path,ca->ca_path,216);
			strcat (path,PATH_DELI);
			strcat (path,"out-");
			strcat (path,lpf->name);
			strcat (path,".crl");

			if(CA_set_CRL(ca)) goto done;
			if(CA_write_crl(ca,path)) goto done;
		}
		break;
	case AICA_OP_CSV:
		if(CA_issue_alotof_certificates(ca,csvf)){
			printf("error in issue_a_lot_of_certificates\n");
			goto done;
		}
		ret++;
		break;
	}

	ret++;
done:
	Cert_free(req);
	return ret;
}

