/* aica_sign.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "ok_asn1.h"
#include "ok_x509.h"
#include "ok_pkcs.h"
#include "ok_tool.h"

#include "ok_aica.h"

Cert *PEM_read_req(char *fname);
void Sign_cert_v3basic(CA *ca,Cert *req);

int ca_ask_yesno(char *txt,int yes);
int ca_ask_num(char *txt,int num);
int confirm_yes_no(char *prompt);
ExtGenNames *input_gennames(char *prompt);

/*-----------------------------------------
  Certificate Authority do Sign.
-----------------------------------------*/
int Sign_do(CA *ca,Cert *req){
	char *cp;
	int i,resig=0,err=-1;

	if(req->issuer) resig=1;

	if(CA_check_policy(ca,req)){
		printf("warning : requested subject isn't matched with CA policy. (%s)\n",req->subject);
		if(!confirm_yes_no("do you continue ? (y/n)[y]: ")) goto done;
	}

	req->version=2;
	if(!resig){
		if(CA_find_statstr(ca,req->subject)){
			printf("warning : requested subject is already used. (%s)\n",req->subject);
			if(!confirm_yes_no("do you continue ? (y/n)[y]: ")) goto done;
		}
		if(!req->serialNumber){
			for(i = ca->cprof->serialNum; 
				CA_find_stat(ca,i) && (ca->cert->serialNumber != i);i++);
			req->serialNumber=i;
			ca->cprof->serialNum=i+1;
		}
		if(Cert_dncopy(&(ca->cert->subject_dn),&(req->issuer_dn))) goto done;
		if((req->issuer=Cert_subject_str(&(req->issuer_dn)))==NULL)  goto done;
	}
	printf("Certificate DATA:\n");
	printf("  serial number : %d\n",req->serialNumber);
	printf("  issuer :\n    %s\n",req->issuer);
	printf("  subject:\n    %s\n",req->subject);

	if(Sign_before_after(ca,req)) goto done;

	if((cp=stm2str(&req->time.notBefore,0))==NULL) goto done;
	printf("  notBefore: %s\n",cp);
	if((cp=stm2str(&req->time.notAfter,0))==NULL) goto done;
	printf("  notAfter : %s\n\n",cp);

	if(!confirm_yes_no("do you sign here ? (y/n)[y]: ")) goto done;
	if(Sign_certext(ca,req)) goto done;

	/* update CA information */
	if(resig){
		if(CA_update_cert(ca,req)) goto done;
	}else{
		if(CA_add_cert(ca,req)) goto done;
	}
	printf("now signing ..\n");

	/* free useless memory first */
	if(req->signature){
		AIFREE(req->signature);
		req->signature = NULL;
	}
	if(req->der) AIFREE(req->der);

	/* get signed certificate */
	OK_set_cert_sig_algo(CA_key2sigalgo(ca->prvkey->key_type));
	if((req->der = Cert_toDER(req,ca->prvkey,NULL,&i))==NULL) goto done;
	err=0;
done:
	return err;
}


int Sign_before_after(CA *ca,Cert *ct){
	time_t t;
	struct tm *stm;
  
	time(&t);

	stm=(struct tm*)gmtime(&t);
	memcpy(&ct->time.notBefore,stm,sizeof(struct tm));

	t+=ca->cprof->cert_days*24*3600;
	if(ct->time.notAfter.tm_year){
		if(mktime(&ca->cert->time.notAfter)<t){
			memcpy(&ct->time.notAfter,&ca->cert->time.notAfter,sizeof(struct tm));
			return 0;}}

	stm=(struct tm*)gmtime(&t);
	memcpy(&ct->time.notAfter,stm,sizeof(struct tm));
	return 0;
}

int Sign_certext(CA *ca,Cert *ct){
	ExtGenNames *egn;
	CertExt *et,*hd=NULL;
	int flg = 0;

	CertExt_free_all(ct->ext);

	/* get new one */
    ct->ext=hd=CertExt_new(-1); /* get dummy */

	for(et=ca->cprof->ext; et ;et=et->next){
		if(et->extnID<=0) continue;

		switch(et->extnID){
		case OBJ_X509v3_AuthKeyIdt:
			flg |= (((CE_AuthKID*)et)->keyid)?(0x4):(0);
			flg |= (((CE_AuthKID*)et)->slen) ?(0x2):(0);
			flg |= (((CE_AuthKID*)et)->authorityCertIssuer)?(0x1):(0);
			if((hd->next=Extnew_authkey_id(ca->cert,flg))==NULL) goto error;
			break;

		case OBJ_X509v3_SbjKeyIdt:
            if((hd->next=Extnew_sbjkey_id(ct))==NULL) goto error;
            break;

		case OBJ_X509v3_SbjAltName:
			if((egn=input_gennames("input SubjectAltName.\n"))==NULL) continue;
			if((hd->next=Extnew_altname(et->extnID,egn))==NULL) goto error;
			break;
		case OBJ_MOJ_RegCoInfo:
			/* pass this one */

		case OBJ_X509v3_BASIC:
		case OBJ_X509v3_NameConst:
		case OBJ_X509v3_PolicyConst:
		case OBJ_X509v3_KEY_Usage:
		case OBJ_X509v3_ExtKeyUsage:
		case OBJ_X509v3_CERT_Pol:
		case OBJ_X509v3_CertPolMap:

		case OBJ_X509v3_IssAltName:
		case OBJ_X509v3_CRL_Point:

		case OBJ_PKIX_IDPE_AIA:
		case OBJ_PKIX_OCSP_NOCHECK:

		case OBJ_NS_CERT_CRLURL:
		case OBJ_NS_CERT_COMMENT:
		case OBJ_NS_CERT_TYPE:

		case OBJ_MOJ_JCertPol:
		case OBJ_MOJ_Registrar:
			if((hd->next=CertExt_dup(et))==NULL) goto error;
			break;
		default:
			continue;
		}
		hd = hd->next;
		hd->critical = et->critical;
	}
	return 0;
error:
	CertExt_free_all(ct->ext);
	ct->ext=NULL;
	return -1;	
}

/*-----------------------------------------
  Confirm yes or no
  output : 1 ... yes
         : 0 ... no.
-----------------------------------------*/
int confirm_yes_no(char *prompt){
	char buf[16];

	printf(prompt);
	fgets(buf,16,stdin);
	if((*buf=='y')||(*buf=='Y')||(*buf=='\n'))
		return 1;

	OK_set_error(ERR_ST_CA_CANCELED,ERR_LC_CA,ERR_PT_CASIGN+3,NULL);
	return 0;
}

/*-----------------------------------------
  Revocate certificate
-----------------------------------------*/
int CA_do_revocation(CA *ca,int snum){
	CertStat *cs;
	int rs=0;

	if((cs=CA_find_stat(ca,snum))==NULL){
		printf("No such certificate.\n");
		return -1;
	}

	printf("------\n");
	printf("Certificate DATA:\n");
	printf("  serial number : %d\n",cs->serialNum);
	printf("  subject:\n    %s\n",cs->subject);
	if(!confirm_yes_no("do you revoke this certificate ? (y/n)[y]: ")) goto error;

	printf("------\n");
	printf("Set revocation reason >>\n");
	printf("  unspecified(0), keyCompromise(1), cACompromise(2),\n");
	printf("  affiliationChanged(3), superseded(4), cessationOfOperation(5),\n");
	printf("  certificateHold(6), removeFromCRL(8), privilegeWithdrawn(9),\n");
	printf("  aaCompromise(10)\n");

	rs = ca_ask_num("select reason code (-1 means 'cancel')",0);

	if(!((rs>=0)&&(rs<=10)&&(rs!=7))) goto error;

	return CA_cert_revoke(ca,snum,rs<<16);
error:
	printf("\nCancel certificate revocation!\n");
	return -1;
}

