/* testserver.c */
/* SSL test server (simple version) */
/*
 * Copyright (C) 1998-2002
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#include <sys/types.h>
#ifdef __WINDOWS__
#undef ULONG
#include <winsock2.h>
#include <io.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include "ok_x509.h"
#include "ok_pkcs.h"
#include "ok_ssl.h"

#define SERVER_PORT 11112

static int myread( int sock,char *buf,int len);
static int mywrite(int sock,char *buf,int len);
static int read_debug(SSL *ssl, int dat);
static int write_debug(SSL *ssl, int dat);
static int vfy_func(SSL *ssl, Cert *ct);

int test_do(SSL *ssl);

int main(int argc, char **argv){
	SSL *s,*ssl;
	Cert *ct;
	int clen;
	char buf[64];
	struct sockaddr_in sin,csin;

#ifdef __WINDOWS__
	WSADATA	wsaData;

	if(WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
		printf("error : WSAStartup\n");
		return(-1);}
#endif

	/* 1. get server socket */
	if (NULL == (s = SSL_socket(AF_INET, SOCK_STREAM, 0))){
		printf("error : SSL_socket\n");
		return(-1);
	}

	memset (&sin,0,sizeof(struct sockaddr_in));
	sin.sin_family      = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port        = htons(SERVER_PORT);          /* Server Port number */

	/* 2. setup server information */
	/* 2.1 read server certificate & private key */
	if(SSL_set_server_p12(s,"00001.p12","abcde")){
		printf("error : SSL_set_server_p12 : %s\n",OK_get_errstr());
		return(-1);
	}

	printf("read server certificate:\n");
	ct = SSL_get_server_cert(s);	/* server PKCS#12 file */
	printf("subject: %s\n",ct->subject);
	printf("issuer:  %s\n",ct->issuer);

	/* 2.2 set certificate request option. */
	printf("set certificate request option.\n");
	SSL_setopt(s,SSL_OPT_CERTREQ);

	/* 2.3 set debug callback functions */
	SSL_set_read_cb(s,myread);
	SSL_set_write_cb(s,mywrite);
	SSL_set_readdebug_cb(s,read_debug);
	SSL_set_writedebug_cb(s,write_debug);
	SSL_set_vfy_cb(s,vfy_func);


	/* 3. bind & listen socket*/
	if(-1 == SSL_bind(s,(struct sockaddr *)&sin,sizeof(sin))){
		printf("error : SSL_bind : %s\n",OK_get_errstr());
		return(-1);
	}
	if(SSL_listen(s,1)){
		printf("error : SSL_listen : %s\n",OK_get_errstr());
		return(-1);
	}

	/* 4. accept the connection */
	clen = sizeof(csin);
	if(NULL == (ssl = SSL_accept(s,(struct sockaddr *)&csin,&clen))){
		printf("error: SSL_accept : %s\n",OK_get_errstr());
		SSL_close(ssl);
		return(-1);
	}

	SSL_close(s);

	/* 5. do SSL handshake :-) */
	if(SSL_handshake(ssl)){
		printf("error: SSL_handshake : %s\n",OK_get_errstr());
		SSL_close(ssl);
		return(-1);
	}

	printf("SSL connection was established!\n");
	SSL_cspec_str(ssl->ctx,buf);
	printf("using cipher : %s\n",buf);

	if(ssl->ctx->cp12){
		printf("read client certificate:\n");
		ct = SSL_get_client_cert(ssl);	/* client PKCS#12 file */
		printf("subject: %s\n",ct->subject);
		printf("issuer:  %s\n",ct->issuer);
	}

	/* 6. and now, send messages with SSL_read & SSL_write !! */
	if(test_do(ssl))
	    return -1;


	/* FREE SSL socket */
	SSL_free(s);
	SSL_free(ssl);

#ifdef __WINDOWS__
	WSACleanup();
#endif
	return 0;
}

int test_do(SSL *ssl){
	unsigned char buf[256];
	int i;

	memset(buf,0,256);
	/* read message first */
	if((i=SSL_read(ssl,buf,256))<0){
		printf("error : SSL_read() : %s\n",OK_get_errstr());
		return -1;
	}
	printf("now getting one message : %s\n",buf);

	strcpy(buf,"I hear you!!");
	printf("now writing one message : %s\n",buf);
	if(SSL_write(ssl,buf,strlen(buf))<0){
		printf("error : SSL_write() : %s\n",OK_get_errstr());
		return -1;
	}

	/* server doesn't need to get close_notify.
	 * this one is just for test.
	 */
	SSL_read(ssl,buf,256);

	SSL_close(ssl);
	return 0;
}

static int read_debug(SSL *ssl, int dat){
    printf("s:r_debug  : md=%d, op=%.2x, st=%d, %d\n",
	    ssl->mode,ssl->opt,ssl->ctx->state,dat);
    return 0;
}

static int write_debug(SSL *ssl, int dat){
    printf("s:w_debug : md=%d, op=%.2x, st=%d, %d\n",
	    ssl->mode,ssl->opt,ssl->ctx->state,dat);
    return 0;
}

static int vfy_func(SSL *ssl, Cert *ct){
    printf("s:vfy_func: %s\n",ct->subject);
    return 0;
}

static int myread(int sock,char *buf,int len){
    int i;
#ifdef __WINDOWS__
	i=recv(sock,buf,len,0);
#else
	i=read(sock,buf,len);
#endif
    printf("&%d&",i);
    return i;
}

static int mywrite(int sock,char *buf,int len){
    int i;
#ifdef __WINDOWS__
	i=send(sock,buf,len,0);
#else
	i=write(sock,buf,len);
#endif
    printf("#%d#",i);
    return i;
}
