/* aistore.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "aiconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ok_asn1.h"
#include "ok_x509.h"
#include "ok_pkcs.h"
#include "ok_pem.h"
#include "ok_tool.h"
#include "ok_store.h"

#include "ok_aica.h"

#ifndef AICONFIG
#define AICONFIG "aica.cnf"
#endif

/* gloval value */
char conf[256] = AICONFIG;
char path[256] = "/";
char pass[32]  = "passwd";
int new   = 0;
int del   = 0;
int print = 0;
int expf  = 0;
int  type = CSTORE_CTX_CERT;
char *sto = STORE_MY;
char *id  = NULL;
char *imp = NULL;
char *expt= NULL;


void options(int argc,char **argv);
void usage(void);
int store_config(char *fname);
#ifdef __WINDOWS__
int get_store_path(char *ret,int max);
#endif

int ai_list_proc(STManager *stm, char *store, int type);
int ai_print_proc(STManager *stm, char *store, int type, char *id);
int ai_delete_proc(STManager *stm, char *store, int type, char *id);
int del_pkey_bag(STManager *stm, CSBag *cbg);
int ai_import_proc(STManager *stm, char *fname);
int update_store_p12(STManager *stm, PKCS12 *p12, char *uid);
int ai_export_proc(STManager *stm, char *store, int type, char *id, char *fname, int format);
Key *get_pkey_bag(STManager *stm, CSBag *cbg);

/*------------------------------------
  main
------------------------------------*/
int main(int argc,char **argv){
  STManager *stm = NULL;

  OK_clear_error();
  options(argc,argv);

#ifdef __WINDOWS__
  if(!get_store_path(path,254)){
     /* no error. continue process */
  }else 
#endif
  if(store_config(conf)){
    printf("cannot find a config file : %s\n",conf);
    goto done;
  }

  /* initialize */
  if(new){
    if((stm=STM_system_new(path))==NULL)
      printf("cannot initialize certificate store : %s\n%s\n",path,OK_get_errstr());
    goto done;
  }

  /* open */
  if((stm=STM_open(path))==NULL){
    printf("cannot open certificate store : %s\n%s\n",path,OK_get_errstr());
    goto done;
  }

  /* do operation */
  if(print){
    if(ai_print_proc(stm,sto,type,id)){
      printf("cannot print a store data : store=%s,type=%s,id=%s\n",sto,
	     (type==CSTORE_CTX_CERT)?("CERT"):("other"),(id)?(id):("NULL"));
      goto done;
    }
    printf("print a store data successfully.\n");
  }else if(imp){
    if(ai_import_proc(stm,imp)){
      printf("cannot import a file to the store : %s\n%s\n",imp,OK_get_errstr());
      goto done;
    }
    printf("import a file to the store successfully.\n");
  }else if(expt){
    if(ai_export_proc(stm,sto,type,id,expt,expf)){
      printf("cannot export a store data : store=%s,type=%s,id=%s : %s\n%s\n",sto,
	     (type==CSTORE_CTX_CERT)?("CERT"):("other"),
	     (id)?(id):("NULL"),expt,OK_get_errstr());
      goto done;
    }
    printf("export a store data successfully.\n");
  }else if(del){
    if(ai_delete_proc(stm,sto,type,id)){
      printf("cannot delete a store data : store=%s,type=%s,id=%s\n%s\n",sto,
	     (type==CSTORE_CTX_CERT)?("CERT"):("other"),
	     (id)?(id):("NULL"),OK_get_errstr());
      goto done;
    }
    printf("delete a store data successfully.\n");
  }else{ /* list store */
    if(ai_list_proc(stm,sto,type)){
      printf("cannot list a certificate store : store=%s,type=%s\n%s\n",sto,
	     (type==CSTORE_CTX_CERT)?("CERT"):("other"),OK_get_errstr());
      goto done;
    }
    printf("list a certificate store successfully.\n");
  }

done:
  if(stm) STM_close(stm);
  if(imp) free(imp);
  if(expt)free(expt);
  if(id ) free(id);
  return 0;
}

/*-----------------------------------------
  usage and option check
-----------------------------------------*/
void options(int argc,char **argv){
  int   i;

#ifdef __WINDOWS__
  OK_set_pwd_dosmode(1);
#endif

  for(i=1;i<argc;i++){
    if(!strcmp("-conf",argv[i])){
      i++;
      if(i<argc) strncpy(conf,argv[i],254);
    }else if(!strcmp("-new",argv[i])){
      new = 1;
    }else if(!strcmp("-l",argv[i])){
    }else if(!strcmp("-p",argv[i])){
      print = 1;
    }else if(!strcmp("-d",argv[i])){
      del = 1;
    }else if(!strcmp("-i",argv[i])){
      i++;
      if(i<argc) imp = strdup(argv[i]);
    }else if(!strcmp("-e",argv[i])){
      i++;
      if(i<argc) expt = strdup(argv[i]);
    }else if(!strcmp("-st",argv[i])){
      i++;
      if(i<argc){
	if(!strcmp("root",argv[i])) sto = STORE_ROOT;
	else if(!strcmp("sub",argv[i])) sto = STORE_MIDCA;
	else if(!strcmp("other",argv[i])) sto = STORE_OTHER;
      }
    }else if(!strcmp("-tp",argv[i])){
      i++;
      if(i<argc){
	if(!strcmp("crl",argv[i])) type = CSTORE_CTX_CRL;
	else if(!strcmp("key",argv[i])) type = CSTORE_CTX_KEY;
	else if(!strcmp("req",argv[i])) type = CSTORE_CTX_CSR;
      }
    }else if(!strcmp("-id",argv[i])){
      i++;
      if(i<argc) id = strdup(argv[i]);
    }else if(!strcmp("-ef",argv[i])){
      i++;
     if(i<argc){
	if(!strcmp("pem",argv[i])) expf = 1;
	else if(!strcmp("pk7",argv[i]))  expf = 2;
	else if(!strcmp("pk12",argv[i])) expf = 3;
      }
    }else if(!strcmp("-pw",argv[i])){
      i++;
      if(i<argc) strcpy(pass,argv[i]);
    }else{
      printf("option error!\n");
      usage();
      exit(0);
    }
  }
}

void usage(void){
  printf("usage: aistore [operation][options]\n");
  printf("[operation]\n");
  printf("\t-new       : initialize certificate store\n");
  printf("\t-l         : list a store data [default]\n");
  printf("\t-i file    : import a file to the store\n");
  printf("\t-e file    : export a cert or a CRL from the store\n");
  printf("\t-p         : print a store data specified with store name & id\n");
  printf("\t-d         : delete a store data specified with store name & id\n");
  printf("[options]\n");
  printf("\t-st store  : specify store name [my(default),other,sub,root]\n");
  printf("\t-tp type   : specify store type [cert(default),crl]\n");
  printf("\t-id id     : specify store data id\n");
  printf("\t-ef format : export format [der(default),pem,pk7,pk12]\n");
  printf("\t-pw passwd : password for key store data\n");
  printf("\t-conf path : set aica config path\n");
}

/*-----------------------------------------
  read store config
-----------------------------------------*/
int store_config(char *fname){
  char *buf=NULL,*p,*cp;
  int err=-1;

  if((buf=conf_get_section(fname,"[general info]","[general info end]"))==NULL)
    goto done;

  if((cp=strstr(buf,"store_dir"))==NULL) goto done;
  if((cp=strchr(cp,'='))==NULL) goto done;
  p=cp+1;

  if(cp=strstr(p,"\r\n")) *cp=0;
  else if(cp=strstr(p,"\n")) *cp=0;
  else goto done;
  strncpy(path,p,254);

  err = 0;
done:
  if(buf) AIFREE(buf);
  return err;
}


/*-----------------------------------------
  list proc
-----------------------------------------*/
int ai_list_proc(STManager *stm, char *store, int type){
  CStore *cs;
  CSBag *bg;
  Cert *ct;
  CRL *crl;
  int i;

  if((cs=STM_find_byName(stm,store,CSTORE_ON_STORAGE,type))==NULL) return -1;

  switch(type){
  case CSTORE_CTX_CERT: printf("[unique-id]\t\tsubject\t\t\tserialNumber\n"); break;
  case CSTORE_CTX_CRL:  printf("[unique-id]\t\tissuer\n"); break;
  }
  printf("----------------------------------------------------------------------\n");

  /* set items into list control */
  for(i=0,bg=cs->bags; bg ; i++,bg=bg->next){
    if(bg->ctx_type == CSTORE_CTX_CERT){
      if((ct=(Cert*)bg->cache)==NULL) continue; /* should be cached */

      printf("[");
      printf(bg->unique_id);
      printf("]  ");
      printf(ct->subject);
      printf("  %d\n",ct->serialNumber);
    }else if(bg->ctx_type == CSTORE_CTX_CRL){
      if((crl=(CRL*)bg->cache)==NULL) continue; /* should be cached */
      printf("[");
      printf(bg->unique_id);
      printf("]  ");
      printf(crl->issuer);
      printf("\n");
    }else{
      printf("[");
      printf(bg->unique_id);
      printf("]\n");
    }
  }
  return 0;
}

/*-----------------------------------------
  list proc
-----------------------------------------*/
int ai_print_proc(STManager *stm, char *store, int type, char *id){
  CSBag *bg;

  if(id==NULL) goto error;
  if((bg=STM_find_byID(stm,store,CSTORE_ON_STORAGE,type,id))==NULL) goto error;

  switch(type){
  case CSTORE_CTX_CERT: Cert_print((Cert*)bg->cache); break;
  case CSTORE_CTX_CRL:  CRL_print((CRL*)bg->cache); break;
  }
  return 0;
error:
  return 1;
}

/*-----------------------------------------
  import proc
-----------------------------------------*/
int ai_import_proc(STManager *stm, char *fname){
  Cert *ct = NULL;
  CRL *crl = NULL;
  Key *key = NULL;
  PKCS7 *p7b = NULL;
  PKCS12 *p12 = NULL;
  int ok = -1;

  OK_clear_passwd();

  /* check input file */
  if(ct=Cert_read_file(fname)){
  }else if(crl=CRL_read_file(fname)){
  }else if(p7b=P7b_read_file(fname)){
  }else if(p12=P12_read_file(fname)){
  }else{
    OK_clear_error();
    goto done;
  }

  OK_clear_error();
  if(STM_reload(stm)) goto done;

  /* update cert store */
  if(ct){
    if(STM_import_cert(stm,ct,id)) goto done;

  }else if(crl){
    if(STM_import_crl(stm,crl,id)) goto done;

  }else if(p7b){
    if(update_store_p12(stm,(PKCS12*)p7b, NULL)) goto done;

  }else if(p12){
    OK_set_passwd(pass);

    if(P12_check_chain(p12,0)) goto done;
    if((ct=P12_get_usercert(p12))==NULL) goto done;
    if((key=P12_get_privatekey(p12))==NULL) goto done;
    if(STM_import_certkey(stm,ct,key,id)) goto done;

    if(update_store_p12(stm,p12,id)) goto done;
  }
  if(STM_update(stm)) goto done;

  ok = 0;
done:
  OK_clear_passwd();
  return ok;
}

int update_store_p12(STManager *stm, PKCS12 *p12, char *uid){
  P12_Baggage *bg;

  /* if same certificate or CRL is already in the Store
   * ignore to install such pkcs12 bag.
   */
  for(bg=p12->bag; bg ;bg=bg->next){
    if(bg->type == OBJ_P12v1Bag_CERT){
      if(STM_find_byCert(stm,((P12_CertBag*)bg)->cert)==NULL){
        /* ignore it's error */
        STM_import_cert(stm,((P12_CertBag*)bg)->cert,uid);
      }
    }else if(bg->type == OBJ_P12v1Bag_CRL){
      if(STM_find_byCRL(stm,((P12_CRLBag*)bg)->crl)==NULL){
        /* ignore it's error */
        STM_import_crl(stm,((P12_CRLBag*)bg)->crl,uid);
      }
    }
  }
  return 0;
}

/*-----------------------------------------
  export proc
-----------------------------------------*/
int ai_export_proc(STManager *stm, char *store, int type, char *id, char *fname, int format){
  CertList *cl,*top=NULL;
  CStore *cs;
  CSBag  *bg = NULL;
  PKCS12 *p12= NULL;
  Key *key=NULL;

  if(id==NULL) goto error;
  if(STM_reload(stm)) goto error;

  /* get store */
  if((cs=STM_find_byName(stm,store,CSTORE_ON_STORAGE,type))==NULL) goto error;

  if((bg=CStore_find_byID(cs->bags,id))==NULL) return 1;

  if(format==3){ /* output PKCS#12 */
    if(bg->cache==NULL) goto error;
    if(strcmp(store,STORE_MY)) goto error;

    if((p12=P12_new())==NULL) goto error;
    if((top=STM_get_pathcert(stm,(Cert*)bg->cache))==NULL) goto error;

    for(cl=top; cl ; cl=cl->next){
      if(P12_add_cert(p12,cl->cert,NULL,0xff)) goto error;
      cl->cert=NULL;
    }

    /* get key */
    if((key=get_pkey_bag(stm,bg))==NULL) goto error;
    if(P12_add_key(p12,key,NULL,0xff)) goto error;

    if(P12_check_chain(p12,0)) goto error;
    if(P12_write_file(p12,fname)) goto error;
    OK_clear_passwd();

    P12_free(p12); Certlist_free_all(top);

  }else if(format==2){ /* output PKCS#7 */
    if(bg->cache==NULL) goto error;
    
    if((p12=P12_new())==NULL) goto error;
    if((top=STM_get_pathcert(stm,(Cert*)bg->cache))==NULL) goto error;

    for(cl=top; cl ; cl=cl->next){
      if(P12_add_cert(p12,cl->cert,NULL,0xff)) goto error;
      cl->cert=NULL;
    }
      
    if(P12_check_chain(p12,0)) goto error;
    if(P7b_write_file((PKCS7*)p12,fname)) goto error;

    P12_free(p12); Certlist_free_all(top);

  }else if(format==1){ /* output a certificate or a CRL (PEM) */
    if(bg->cache==NULL) goto error;
    if(type==CSTORE_CTX_CERT){
      if(PEM_write_cert((Cert*)bg->cache,fname)) goto error;
    }else{
      if(PEM_write_crl((CRL*)bg->cache,fname)) goto error;
    }

  }else{ /* output a certificate or a CRL (DER) */
    if(ASN1_write_der(bg->der,fname)) goto error;
  }

  return 0;
error:
  return -1;
}

Key *get_pkey_bag(STManager *stm, CSBag *cbg){
  CStore *ks;
  CSBag *kbg;
  Key *ret = NULL;
  char buf[32];

  if((ks=STM_find_byName(stm,STORE_MY,CSTORE_ON_STORAGE,CSTORE_CTX_KEY))==NULL) goto done;

  /* check private key password */
  OK_set_passwd(pass);

  if((kbg=CStore_find_byKeyHash(ks->bags,cbg->key_hash))==NULL) goto done;

  /* first time */
  if(ret=CStore_get_key(ks,kbg)) goto done;

  /* second time */
  OK_get_passwd("Access Private Key:",(unsigned char*)buf,0);
  OK_set_passwd(buf);
  if(ret=CStore_get_key(ks,kbg)) goto done;

done:
  OK_clear_passwd();
  memset(buf,0,32);
  return ret;
}


/*-----------------------------------------
  delete proc
-----------------------------------------*/
int ai_delete_proc(STManager *stm, char *store, int type, char *id){
  CStore *cs;
  CSBag *bg;

  if(id==NULL) goto error;
  if(STM_reload(stm)) goto error;

  /* get key hash */
  if((cs=STM_find_byName(stm,store,CSTORE_ON_STORAGE,type))==NULL) goto error;

  if((bg=CStore_find_byID(cs->bags,id))==NULL) return 1;

  if(!strcmp(store,STORE_MY)){ 
    if(del_pkey_bag(stm,bg)) goto error;
  }

  /* clean certificate bag */
  if(CStore_del_byID(cs,id)) goto error;

  /* update store */
  if(STM_update(stm)) goto error;

  return 0;
error:
  return -1;
}

int del_pkey_bag(STManager *stm, CSBag *cbg){
  CStore *ks;
  CSBag *kbg;
  Key *tmp;
  char buf[32];

  if((ks=STM_find_byName(stm,STORE_MY,CSTORE_ON_STORAGE,CSTORE_CTX_KEY))==NULL) goto error;

  /* check private key password */
  OK_set_passwd(pass);

  if((kbg=CStore_find_byKeyHash(ks->bags,cbg->key_hash))==NULL) goto error;
  if((tmp=CStore_get_key(ks,kbg))==NULL){
    OK_get_passwd("Access Private Key:",(unsigned char*)buf,0);
    OK_set_passwd(buf);
    if((tmp=CStore_get_key(ks,kbg))==NULL) goto error;
  }
  Key_free(tmp);
  OK_clear_passwd();
  memset(buf,0,32);

  if(CStore_del_bag(ks,kbg)) goto error;

  return 0;
error:
  return -1;
}

