/* ca_crl.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "ok_asn1.h"
#include "ok_rsa.h"
#include "ok_x509.h"
#include "ok_pkcs.h"
#include "ok_tool.h"

#include "ok_ca.h"

/*-----------------------------------------
  revoke a certificate with serial num.
-----------------------------------------*/
int CA_cert_revoke(CA *ca,int revoke,int reason){
	CertStat *st;
	time_t t;
	struct tm *stm;

	if((st=CA_find_stat(ca,revoke))==NULL) return -1;

	st->state |= STAT_REVOKED | (reason&0x00ff0000);

	time(&t);

	stm=(struct tm*)gmtime(&t);
	memcpy(&st->revokedDate,stm,sizeof(struct tm));

	return 0;
}


/*-----------------------------------------
  create new CRL
-----------------------------------------*/
int CA_set_CRL(CA *ca){
	CRL *crl;
	int i;

	if(ca->crl) CRL_free(ca->crl);
	if((crl=ca->crl=CRL_new())==NULL) goto error;

	/* set CRL signature algorithm */
	OK_set_crl_sig_algo(ca->lprof->sigtype);

	if(Cert_dncopy(&(ca->cert->subject_dn),&(crl->issuer_dn))) goto error;
	if((crl->issuer= Cert_subject_str(&(crl->issuer_dn)))==NULL) goto error;

	if(CA_crl_last_next(ca,crl)) goto error;
	if(CA_crl_set_revoked(ca,crl,&i)) goto error;

	if(ca->lprof->crl_ver)
		if(CA_crl_set_ext(ca,crl)) goto error;

	OK_set_crl_sig_algo(CA_key2sigalgo(ca->prvkey->key_type));
	if((crl->der=CRL_toDER(crl,ca->prvkey,NULL,&i))==NULL) goto error;

	return 0;
error:
	CRL_free(crl);
	ca->crl=NULL;
	return -1;
}


int CA_crl_last_next(CA *ca,CRL *crl){
	time_t t;
	struct tm *stm;

	time(&t);

	stm = gmtime(&t);
	memcpy(&crl->lastUpdate,stm,sizeof(struct tm));

	t+=ca->lprof->crl_days*24*3600;

	stm = gmtime(&t);
	memcpy(&crl->nextUpdate,stm,sizeof(struct tm));
	return 0;
}

int CA_crl_set_revoked(CA *ca,CRL *crl,int *rev_num){
	CertProf *cpf;
	CertStat *st;
	Revoked  *rv,*tmp;
	int type=0,rsn=0;

	if(!strcmp(ca->lprof->name,"CRL")) type = 1; /* only contain User cert */
	if(!strcmp(ca->lprof->name,"ARL")) type = 2; /* only contain CA cert */

	rsn = (CertExt_find(ca->lprof->ext,OBJ_X509v3_CRLReason)!=NULL);

	*rev_num=0;
	for(cpf=ca->profList; cpf ;cpf=cpf->next){
		for(st=cpf->stat; st ;st=st->next){
			if(st->state&STAT_REVOKED){
				if((type==1)&&(st->state&(STAT_CA|STAT_XCERTFWD))) continue;
				if((type==2)&&(!(st->state&(STAT_CA|STAT_XCERTFWD)))) continue;

				if((rv=Revoked_new())==NULL) goto error;
				rv->serialNumber = st->serialNum;
				memcpy(&rv->revocationDate,&st->revokedDate,sizeof(struct tm));

				if(rsn){
					if((rv->entExt=Extnew_reason_code((st->state>>16)&0xff)) == NULL)
						goto error;
				}

				tmp=crl->next;
				crl->next=rv;
				rv->next =tmp;
				(*rev_num)++;
			}
		}
	}
	return 0;
error:
	return -1;
}

int CA_crl_set_ext(CA *ca,CRL *crl){
	CertExt *et,*hd=NULL;
	int i,flg = 0;

    CertExt_free_all(crl->ext);
    crl->ext=hd=CertExt_new(-1); /* get dummy */

	for(et=ca->lprof->ext; et ;et=et->next){
		if(et->extnID<=0) continue;

		switch(et->extnID){
		case OBJ_X509v3_AuthKeyIdt:
			flg |= (((CE_AuthKID*)et)->keyid)?(0x4):(0);
			flg |= (((CE_AuthKID*)et)->slen) ?(0x2):(0);
			flg |= (((CE_AuthKID*)et)->authorityCertIssuer)?(0x1):(0);
			if((hd->next=Extnew_authkey_id(ca->cert,flg))==NULL) goto error;
			break;
		case OBJ_X509v3_CRLNumber:
			if((hd->next=Extnew_crl_number(ca->lprof->crlNum))==NULL) goto error;
			i = (ca->lprof->crlNum+1)%10000000;
			ca->lprof->crlNum = (i>0)?(i):(1);
			break;
		case OBJ_X509v3_CRLReason:
			/* just flag for entry extension */
			continue;
		case OBJ_X509v3_IssDistPoint:
		case OBJ_X509v3_IssAltName:
		default:
			if((hd->next=CertExt_dup(et))==NULL) goto error;
			break;
		}
		hd = hd->next;
		hd->critical = et->critical;
	}

	return 0;
error:
	return -1;
}
