/* ca_tool.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ok_x509.h"
#include "ok_x509ext.h"
#include "ok_asn1.h"
#include "ok_pkcs.h"
#include "ok_pem.h"

#include "ok_ca.h"

extern char path[];
extern char caf[];
extern char info[];

/*-----------------------------------------
  count issued certificates
-----------------------------------------*/
int CA_now_certnum(CA *ca){
	CertProf *cpf;
	int ret=0;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		ret+=Prof_now_certnum(cpf);

	return ret;
}

/*-----------------------------------------
  is the serial num usable ?
  output : 0 ... no problem to use
         : 1 ... it's already used
         : -1... system error
-----------------------------------------*/
int CA_can_serialNum(CA *ca,Cert *req,int sn){
	CertStat *st;

	if(ca->cert->serialNumber == sn){
		OK_set_error(ERR_ST_CA_BADSN,ERR_LC_CA,ERR_PT_CATOOL+2,NULL);
		return -1;
	}

	if(st=CA_find_stat(ca,sn)){
		if(!(st->state&STAT_EXPIRED)){
			OK_set_error(ERR_ST_CA_BADSN,ERR_LC_CA,ERR_PT_CATOOL+2,NULL);
			return 1;
		}
		CA_delete_stat(st);
	}
	req->serialNumber = sn;
	return 0;
}

/*-----------------------------------------
  is certificate possible to resign ?
  output : 0 ... no problem
         : 1 ... not possible
         : -1... system error;
-----------------------------------------*/
int CA_can_resign(CA *ca,Cert *ct){
	CertStat *st;

	if(strcmp(ca->subject,ct->issuer)){
		OK_set_error(ERR_ST_CA_BADSUBJECT,ERR_LC_CA,ERR_PT_CATOOL+3,NULL);
		return 1;
	}

	if((st=CA_find_stat(ca,ct->serialNumber))==NULL) return 1;
	if((ca->cprof=CA_find_prof_sn(ca,ct->serialNumber))==NULL) return 1;

	if(st->state&STAT_REVOKED){
		OK_set_error(ERR_ST_CA_CTREVOKED,ERR_LC_CA,ERR_PT_CATOOL+3,NULL);
		return 1;
	}

	return 0;
}

/*-----------------------------------------
  check expiration of certificates
  output : 0 ... no expired cert
         : 1 ... some certs was expired
-----------------------------------------*/
int CA_expire_check(CA *ca){
	CertProf *cpf;
	int ret=0;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		ret+=Prof_expire_check(cpf);

	return ret;
}

/*-----------------------------------------
  CA add new certificate info
-----------------------------------------*/
int CA_add_cert(CA *ca,Cert *ct){
	return Prof_add_cert(ca->cprof,ct);
}

int CA_add_certfile(CA *ca,Cert *ct){
	return Prof_add_certfile(ca->cprof,ct);
}

/*-----------------------------------------
  CA delete certificate info
-----------------------------------------*/
void CA_delete_stat(CertStat *st){
	/* we don't need "(ca->stat=st->next);" */
	Prof_delete_stat(st);
}

/*-----------------------------------------
  CA update certificate info
-----------------------------------------*/
int CA_update_cert(CA *ca,Cert *ct){
	CertStat *st;

	if((st=CA_find_stat(ca,ct->serialNumber))==NULL) return -1;
	st->state = STAT_OK |((Cert_is_CA(ct)>0)?(STAT_CA):(0));
	memcpy(&st->notBefore,&ct->time.notBefore,sizeof(struct tm));
	memcpy(&st->notAfter,&ct->time.notAfter,sizeof(struct tm));
	memset(&st->revokedDate,0,sizeof(struct tm));
	return 0;
}

/*-----------------------------------------
  CA utility : signature algo from prvkey
-----------------------------------------*/
int CA_key2sigalgo(int type){
	switch(type){
	case KEY_RSA_PRV:
	case KEY_RSA_PUB:
		return OBJ_SIG_SHA1RSA;
	case KEY_DSA_PRV:
	case KEY_DSA_PUB:
		return OBJ_SIG_SHA1DSA;
	case KEY_ECDSA_PRV:
	case KEY_ECDSA_PUB:
		return OBJ_SIG_SHA1ECDSA;
	}
	return -1;
}

/*-------------------------------------------
  CA utility : find CertStat by serial num
-------------------------------------------*/
CertStat *CA_find_stat(CA *ca,int serial){
	CertProf *cpf;
	CertStat *st;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		if(st=Prof_find_stat(cpf,serial))
			return st;

	return NULL;
}

/*-----------------------------------------
  CA utility : find CertStat by subject
-----------------------------------------*/
CertStat *CA_find_statstr(CA *ca,char *subject){
	CertProf *cpf;
	CertStat *st;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		if(st=Prof_find_statstr(cpf,subject))
			return st;

	return NULL;
}

/*-----------------------------------------
  CA utility : find Profile functions
-----------------------------------------*/
CertProf *CA_find_prof_sn(CA *ca,int serial){
	CertProf *cpf;
	CertStat *st;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		if(st=Prof_find_stat(cpf,serial))
			return cpf;
	return NULL;
}

CertProf *CA_find_prof_sbj(CA *ca,char *subject){
	CertProf *cpf;
	CertStat *st;

	for(cpf=ca->profList; cpf; cpf=cpf->next)
		if(st=Prof_find_statstr(cpf,subject))
			return cpf;
	return NULL;
}
