/* prof_tool.c */
/*
 * Copyright (C) 1998-2001
 * Akira Iwata & Takuto Okuno
 * Akira Iwata Laboratory,
 * Nagoya Institute of Technology in Japan.
 *
 * All rights reserved.
 *
 * This software is written by Takuto Okuno(usapato@anet.ne.jp)
 * And if you want to contact us, send an email to Kimitake Wakayama
 * (wakayama@elcom.nitech.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software must
 *    display the following acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *    Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Akira Iwata Laboratory,
 *     Nagoya Institute of Technology in Japan (http://mars.elcom.nitech.ac.jp/)."
 *
 *   THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
 *   AKIRA IWATA LABORATORY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 *   IN NO EVENT SHALL AKIRA IWATA LABORATORY BE LIABLE FOR ANY SPECIAL,
 *   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 *   FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 *   NEGLIGENCE OR OTHER TORTUOUS ACTION, ARISING OUT OF OR IN CONNECTION
 *   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ok_asn1.h"
#include "ok_x509.h"
#include "ok_x509ext.h"
#include "ok_pkcs.h"
#include "ok_pem.h"

#include "ok_ca.h"

/*-----------------------------------------
  count issued certificates
-----------------------------------------*/
int Prof_now_certnum(CertProf *cpf){
	CertStat *st;
	int ret=0;

	for(st=cpf->stat;st;st=st->next) ret++;
	return ret;
}

/*-----------------------------------------
  count extensions
-----------------------------------------*/
int Prof_now_extnum(CertProf *cpf){
	CertExt *et=cpf->ext;
	int i=0;

	while(et){ i++; et=et->next; }
	return i;
}

/*-----------------------------------------
  check expiration of certificates
  output : 0 ... no expired cert
         : 1 ... some certs was expired
-----------------------------------------*/
int Prof_expire_check(CertProf *cpf){
	CertStat *st;
	time_t now,t;
	int ret=0;

	time(&now); /* get current time */

	for(st=cpf->stat; st ;st=st->next){
		/* notAfter */
		t=mktime(&st->notAfter);
		if(now>t){ st->state |= STAT_EXPIRED; ret=1; }

		/* notBefore */
		t=mktime(&st->notBefore);
		if(now<t){ st->state |= STAT_EXPIRED; ret=1; }
	}
	return ret;
}

/*-------------------------------------------
  CA utility : find CertStat by serial num
-------------------------------------------*/
CertStat *Prof_find_stat(CertProf *cpf,int serial){
	CertStat *st;
	for(st=cpf->stat;st;st=st->next)
		if(st->serialNum==serial)
			return st;

	OK_set_error(ERR_ST_CA_NOCSTAT,ERR_LC_PROF,ERR_PT_PFTOOL+2,NULL);
	return NULL;
}

/*-----------------------------------------
  CA utility : find CertStat by subject
-----------------------------------------*/
CertStat *Prof_find_statstr(CertProf *cpf,char *subject){
	CertStat *st;
	for(st=cpf->stat;st;st=st->next)
		if(!strcmp(st->subject,subject))
			return st;

	OK_set_error(ERR_ST_CA_NOCSTAT,ERR_LC_PROF,ERR_PT_PFTOOL+3,NULL);
	return NULL;
}

CertProf *Prof_find(CA *ca,char *name){
	CertProf *cpf;
	for(cpf=ca->profList; cpf ; cpf=cpf->next)
		if(!strcmp(cpf->name,name))
			return cpf;

	OK_set_error(ERR_ST_CA_NOPROF,ERR_LC_PROF,ERR_PT_PFTOOL+4,NULL);
	return NULL;
}

CRLProf *Prof_crl_find(CA *ca,char *name){
	CRLProf *cpf;
	for(cpf=ca->crlpList; cpf ; cpf=cpf->next)
		if(!strcmp(cpf->name,name))
			return cpf;

	OK_set_error(ERR_ST_CA_NOPROF,ERR_LC_PROF,ERR_PT_PFTOOL+4,NULL);
	return NULL;
}

/*---------------------------------------------------------------*/
/*-----------------------------------------
  CA add new certificate info
-----------------------------------------*/
int Prof_add_cert(CertProf *cpf,Cert *ct){
	CertStat *st,*tmp;

	if((st = CertStat_new())==NULL) goto error;
	st->state = STAT_OK |((Cert_is_CA(ct)>0)?(STAT_CA):(0));
	st->serialNum = ct->serialNumber;

	if((STRDUP(st->subject,ct->subject))==NULL){
		OK_set_error(ERR_ST_MEMALLOC,ERR_LC_PROF,ERR_PT_PFTOOL+5,NULL);
		goto error;		
	}
	memcpy(&st->notBefore,&ct->time.notBefore,sizeof(struct tm));
	memcpy(&st->notAfter, &ct->time.notAfter, sizeof(struct tm));
	memset(&st->revokedDate,0,sizeof(struct tm));

	tmp=cpf->stat;
	st->next=tmp;
	st->prev=NULL;
	cpf->stat=st;
	if(tmp) tmp->prev=st;
	return 0;
error:
	CertStat_free(st);
	return -1;
}

/*-----------------------------------------
  CA delete certificate info
-----------------------------------------*/
void Prof_delete_stat(CertStat *st){
	/* we don't need "(cpf->stat=st->next);" */
	if(st==NULL) return;
	if(st->prev) st->prev->next=st->next;
	if(st->next) st->next->prev=st->prev;

	CertStat_free(st);
}

/*-----------------------------------------
  CA add new certificate file
-----------------------------------------*/
int Prof_add_certfile(CertProf *cpf,Cert * ct){
	CertStat *st;
	unsigned char buf[128];
	int	i;

	if((st = Prof_find_stat(cpf,ct->serialNumber))==NULL) return -1;

	ASN1_set_integer(ct->serialNumber,buf,&i);	/* set serialNum (for reqInfo) */
	if(fsetpos(cpf->fp,&(cpf->pos))) goto error;
	if(fwrite(buf,sizeof(char),i,cpf->fp)<(unsigned)i) goto error;

	st->fp = cpf->fp;
	if(fgetpos(cpf->fp,&(st->pos))) goto error;

	if(ASN1_skip_(ct->der,&i)==NULL) goto error;
	if(fwrite(ct->der,sizeof(char),i,cpf->fp)<(unsigned)i) goto error;
	if(fgetpos(cpf->fp,&(cpf->pos))) goto error;

	/* commit to disk */
	if(fflush(cpf->fp)) goto error;

	return 0;
error:
	OK_set_error(ERR_ST_FILEWRITE,ERR_LC_PROF,ERR_PT_PFTOOL+6,NULL);
	return -1;
}

int Prof_add_xcertfile(CertProf *cpf,CertPair * cp){
	CertStat *st;
	unsigned char buf[128];
	char *sbj;
	int	i;

	if(cp->issuedByThisCA){
		sbj = cp->issuedByThisCA->subject;
		if((st = Prof_find_stat(cpf,cp->issuedByThisCA->serialNumber))==NULL) goto error;
	}else if(cp->issuedToThisCA){
		sbj = cp->issuedToThisCA->issuer;
		if((st = Prof_find_statstr(cpf,cp->issuedToThisCA->issuer))==NULL) goto error;
	}

	ASN1_set_integer(st->serialNum,buf,&i);	/* set serialNum (for reqInfo) */
	if(fsetpos(cpf->fp,&(cpf->pos))) goto error;
	if(fwrite(buf,sizeof(char),i,cpf->fp)<(unsigned)i) goto error;

	st->fp = cpf->fp;
	if(fgetpos(cpf->fp,&(st->pos))) goto error;

	if(ASN1_skip_(cp->der,&i)==NULL) goto error;
	if(fwrite(cp->der,sizeof(char),i,cpf->fp)<(unsigned)i) goto error;
	if(fgetpos(cpf->fp,&(cpf->pos))) goto error;

	/* commit to disk */
	if(fflush(cpf->fp)) goto error;

	return 0;
error:
	OK_set_error(ERR_ST_FILEWRITE,ERR_LC_PROF,ERR_PT_PFTOOL+7,NULL);
	return -1;
}

/*---------------------------------------------------------------*/
/*-----------------------------------------
  add CertExt to the Prof structure
-----------------------------------------*/
int Prof_add_ext(CertProf *cpf,CertExt *et){
	CertExt *hd=NULL;

	if(et==NULL){
		OK_set_error(ERR_ST_NULLPOINTER,ERR_LC_PROF,ERR_PT_PFTOOL+8,NULL);
		return -1;
	}

	for(hd=cpf->ext; hd ; hd=hd->next)
		if(hd->next == NULL) break;

	if(hd) hd->next = et;
	else   cpf->ext = et;

	return 0;
}

/*-----------------------------------------
  delete any CertExt in Prof structure
-----------------------------------------*/
int Prof_delete_ext(CertProf *cpf,int id){
	CertExt *et,*pv;

	for(et=cpf->ext;et;et=et->next){
		if(et->extnID==id) break;
		pv = et;
	}

	if(et==NULL){
		OK_set_error(ERR_ST_MEMALLOC,ERR_LC_PROF,ERR_PT_PFTOOL+9,NULL);
		return -1;
	}

	if(et == cpf->ext){
		cpf->ext = et->next;
	}else{
		pv->next = et->next;
	}
	CertExt_free(et);
	return 0;
}
